// XFile.cpp: implementation of the XFile class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include <rmxftmpl.h>
#include <initguid.h>
#include "..\INCLUDE\XFile.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

//	error macros I use, sorry no sorce 8p
//	replace with anything you like
#define D3DEX(_h_result_)		_h_result_
#define PEASSERT(_eval_)

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

XFile::XFile()
: m_pCurrentType(NULL)
{
}

XFile::~XFile()
{
}

///////////////////////////////////////////////////////////////////////////////
// Function name	: XFile::Create
// Description	    : Initializes the XFile object and sets up iteration
// Return type		: void 
// Argument         : void *pvSource   :pointer to source file name or raw data
// Argument         : void *pvTemplate :pointer to template file name or data
// Argument         : DWORD dwTemplateSize :size of template data in bytes
// Argument         : DWORD dwLoadFlags :pvSource data type flag
// Argument         : bool bLoadTemplatesFromFile :pvTemplate data type flag
//
void XFile::Create(void *pvSource, const void *pvTemplate /*= NULL*/,
				   DWORD dwTemplateSize /*= 0*/,
				   DWORD dwLoadFlags /*= DXFILELOAD_FROMFILE*/,
				   bool bLoadTemplatesFromFile /*= false*/,
				   bool bUseStandardTemplates /*= true*/)
{
	IDirectXFilePtr				pXFile;
	IDirectXFileEnumObjectPtr	pEnumObject;

//	Create the XFile object interface
	D3DEX(DirectXFileCreate(&pXFile.Interface()));

//	Register the standard templates
	if(bUseStandardTemplates)
		D3DEX(pXFile->RegisterTemplates((VOID*)D3DRM_XTEMPLATES,
										D3DRM_XTEMPLATE_BYTES));

//	If there are custom templates try to load them
	if(NULL != pvTemplate)
	{
		if(bLoadTemplatesFromFile)
		{
// Not supported yet, pvTemplate should be file name, path, stream or something
			PEASSERT(bLoadTemplatesFromFile == true);
		}
		else
		{
			D3DEX(pXFile->RegisterTemplates(const_cast<void*>(pvTemplate),
											dwTemplateSize));
		}
	}

//	Now create the enum object
	D3DEX(pXFile->CreateEnumObject(pvSource, dwLoadFlags,
								   &pEnumObject.Interface()))
	m_pXFile = pXFile;
	m_pEnumObject = pEnumObject;
}

///////////////////////////////////////////////////////////////////////////////
// Function name	: XFile::Push
// Description	    : Moves iteration to children of current opbjec
// Return type		: bool 
//
bool XFile::Push()
{
	bool bRet = false;

	if(m_pCurrentObject)
	{
		m_stackFileData.push_back(m_pCurrentObject);
		m_pCurrentType = NULL;
		m_pCurrentObject = NULL;
		bRet = true;
	}

	return bRet;
}

///////////////////////////////////////////////////////////////////////////////
// Function name	: XFile::Pop
// Description	    : Moves iteration to parent of current object
// Return type		: bool 
//
bool XFile::Pop()
{
	bool bRet = false;
	
	if(!m_stackFileData.empty())
	{
		m_pCurrentObject = m_stackFileData.back();
		m_stackFileData.pop_back();
		D3DEX(m_pCurrentObject->GetType(&m_pCurrentType));
		bRet = true;
	}

	return bRet;
}

///////////////////////////////////////////////////////////////////////////////
// Function name	: XFile::NextSibling
// Description	    : Moves iteration to next sibling object
// Return type		: bool 
//
bool XFile::NextSibling()
{
	bool bRet = false;

	if(!m_stackFileData.empty())
	{
		IDirectXFileDataPtr pChild;

		if(GetChildDataObject(pChild, m_stackFileData.back()))
		{
			m_pCurrentObject = pChild;
			D3DEX(m_pCurrentObject->GetType(&m_pCurrentType));
			bRet = true;
		}
	}
	return bRet;
}

///////////////////////////////////////////////////////////////////////////////
// Function name	: XFile::NextRoot
// Description	    : Moves iteration to next root object
// Return type		: bool 
//
bool XFile::NextRoot()
{
	IDirectXFileDataPtr pRoot;
	HRESULT	hr;

	m_stackFileData.clear();
	m_pCurrentObject = NULL;
	m_pCurrentType = NULL;

	if(SUCCEEDED(hr = m_pEnumObject->GetNextDataObject(&pRoot.Interface())))
	{
		if(FAILED(hr = pRoot.QueryInterface(m_pCurrentObject)))
		{
			IDirectXFileDataReferencePtr	pReference;
			if(SUCCEEDED(hr = pRoot.QueryInterface(pReference)))
			{
				D3DEX(pReference->Resolve(&m_pCurrentObject.Interface()));
				pReference = NULL;
			}
		}

	}

	if(SUCCEEDED(hr))
	{
		m_pCurrentObject->GetType(&m_pCurrentType);
	}
	return SUCCEEDED(hr);
}

///////////////////////////////////////////////////////////////////////////////
// Function name	: XFile::IsCurrent
// Description	    : Compares type of current object
// Return type		: bool 
// Argument         : REFGUID rGuid
//
bool XFile::IsCurrent(REFGUID rGuid)
{
	PEASSERT(m_pCurrentType);
	return NULL != m_pCurrentType && IsEqualGUID(rGuid, *m_pCurrentType);
}

///////////////////////////////////////////////////////////////////////////////
// Function name	: XFile::CurrentType
// Description	    : Gets the current objects GUID
// Return type		: bool 
// Argument         : GUID & refGuid
//
bool XFile::CurrentType(GUID & refGuid)
{
	if(m_pCurrentType)
		refGuid = *m_pCurrentType;

	return NULL != m_pCurrentType;
}

///////////////////////////////////////////////////////////////////////////////
// Function name	: XFile::CurrentID
// Description	    : Gets the current objects ID
// Return type		: bool 
// Argument         : GUID &refGuid
//
bool XFile::CurrentID(GUID &refGuid)
{
	if(!m_pCurrentObject)
		D3DEX(m_pCurrentObject->GetId(&refGuid));

	return !m_stackFileData.empty();
}

///////////////////////////////////////////////////////////////////////////////
// Function name	: XFile::CurrentName
// Description	    : Gets the current objects name
// Return type		: bool 
// Argument         : std::string &strName
//
bool XFile::CurrentName(std::string &strName)
{
	DWORD	dwLen;
	if(m_pCurrentObject)
	{
		m_pCurrentObject->GetName(NULL, &dwLen);
		strName.resize(dwLen);
		D3DEX(m_pCurrentObject->GetName(&strName[0], &dwLen));
	}
	return 0 != strName.size();
}

IDirectXFileDataPtr& XFile::CurrentDataObject(void)
{
	return m_pCurrentObject;
}

///////////////////////////////////////////////////////////////////////////////
// Function name	: XFile::GetString
// Description	    : Gets a string value
// Return type		: bool 
// Argument         : LPCSTR szName
// Argument         : char *&pszName
//
bool XFile::GetString(LPCSTR szName, char *&pszName)
{
	DWORD	dwSize = 0;
	char	**ppString;
	HRESULT	hr;

	if(SUCCEEDED( hr = m_pCurrentObject->GetData(szName, &dwSize,
									reinterpret_cast<void**>(&ppString))) )
		pszName = *ppString;

	PEASSERT(SUCCEEDED(hr));
	return SUCCEEDED(hr);
}

///////////////////////////////////////////////////////////////////////////////
// Function name	: XFile::GetStringArray
// Description	    : Gets an array of string values
// Return type		: bool 
// Argument         : LPCSTR szName
// Argument         : char **&ppszName
//
bool XFile::GetStringArray(LPCSTR szName, char **&ppszName)
{
	DWORD	dwSize = 0;
	char	**ppString;
	HRESULT	hr = m_pCurrentObject->GetData(szName, &dwSize,
									reinterpret_cast<void**>(&ppString));
	ppszName = ppString;
	PEASSERT(SUCCEEDED(hr));
	return SUCCEEDED(hr);
}

///////////////////////////////////////////////////////////////////////////////
// Function name	: XFile::GetDword
// Description	    : Gets a DWORD value
// Return type		: bool 
// Argument         : LPCSTR szName
// Argument         : DWORD &dwData
//
bool XFile::GetDword(LPCSTR szName, DWORD &dwData)
{
	DWORD	*pDWORD;
	DWORD	dwSize;
	HRESULT	hr = m_pCurrentObject->GetData(szName, &dwSize,
										reinterpret_cast<void**>(&pDWORD));
	if(SUCCEEDED(hr) && sizeof(DWORD) == dwSize)
		dwData = *pDWORD;
	else
		PEASSERT(SUCCEEDED(hr) && sizeof(DWORD) == dwSize);

	return SUCCEEDED(hr);
}

///////////////////////////////////////////////////////////////////////////////
// Function name	: XFile::GetDwordArray
// Description	    : Gets an array of DWORD values
// Return type		: bool 
// Argument         : LPCSTR szName
// Argument         : DWORD *&pDWArray
//
bool XFile::GetDwordArray(LPCSTR szName, DWORD *&pDWArray)
{
	DWORD	dwSize;
	HRESULT	hr = m_pCurrentObject->GetData(szName, &dwSize,
										reinterpret_cast<void**>(&pDWArray));
	PEASSERT(SUCCEEDED(hr) && 0 == dwSize % sizeof(DWORD));
	return SUCCEEDED(hr);
}

///////////////////////////////////////////////////////////////////////////////
// Function name	: XFile::GetFloat
// Description	    : Gets a float value
// Return type		: bool 
// Argument         : LPCSTR szName
// Argument         : float &fData
//
bool XFile::GetFloat(LPCSTR szName, float &fData)
{
	float	*pFloat;
	DWORD	dwSize;
	HRESULT	hr = m_pCurrentObject->GetData(szName, &dwSize,
										reinterpret_cast<void**>(&pFloat));
	if(SUCCEEDED(hr) && sizeof(float) == dwSize)
		fData = *pFloat;
	else
		PEASSERT(SUCCEEDED(hr) && sizeof(float) == dwSize);

	return SUCCEEDED(hr);
}

///////////////////////////////////////////////////////////////////////////////
// Function name	: XFile::GetFloatArray
// Description	    : Gets an array of float values
// Return type		: bool 
// Argument         : LPCSTR szName
// Argument         : float *&pFArray
//
bool XFile::GetFloatArray(LPCSTR szName, float *&pFArray)
{
	DWORD	dwSize;
	HRESULT	hr = m_pCurrentObject->GetData(szName, &dwSize,
										reinterpret_cast<void**>(&pFArray));
	PEASSERT(SUCCEEDED(hr) && 0 == dwSize % sizeof(float));
	return SUCCEEDED(hr);
}

///////////////////////////////////////////////////////////////////////////////
// Function name	: XFile::GetVector3
// Description	    : Gets a 3 component vector
// Return type		: bool 
// Argument         : LPCSTR szName
// Argument         : D3DVECTOR &vecVector
//
bool XFile::GetVector3(LPCSTR szName, D3DVECTOR &vecVector)
{
	D3DVECTOR *pVec;
	DWORD	dwSize;
	HRESULT	hr = m_pCurrentObject->GetData(szName, &dwSize,
											reinterpret_cast<void**>(&pVec));
	PEASSERT(SUCCEEDED(hr) && dwSize ==  sizeof(D3DVECTOR));
	vecVector = *pVec;
	return SUCCEEDED(hr);
}

///////////////////////////////////////////////////////////////////////////////
// Function name	: XFile::GetVector3Array
// Description	    : Gets an array of 3 component vectors
// Return type		: bool 
// Argument         : LPCSTR szName
// Argument         : D3DVECTOR *&pVectors
//
bool XFile::GetVector3Array(LPCSTR szName, D3DVECTOR *&pVectors)
{
	DWORD	dwSize;
	HRESULT	hr = m_pCurrentObject->GetData(szName, &dwSize,
										reinterpret_cast<void**>(&pVectors));
	PEASSERT(SUCCEEDED(hr) && 0 == dwSize % sizeof(D3DVECTOR));
	return SUCCEEDED(hr);
}

///////////////////////////////////////////////////////////////////////////////
// Function name	: XFile::GetColorRGBA
// Description	    : Gets an RGBA color value
// Return type		: bool 
// Argument         : LPCSTR szName
// Argument         : D3DCOLORVALUE &rgbColorRGBA
//
bool XFile::GetColorRGBA(LPCSTR szName, D3DCOLORVALUE &rgbColorRGBA)
{
	D3DCOLORVALUE *pColor;
	DWORD	dwSize;
	HRESULT	hr = m_pCurrentObject->GetData(szName, &dwSize,
										reinterpret_cast<void**>(&pColor));
	PEASSERT(SUCCEEDED(hr) && dwSize == sizeof(D3DCOLORVALUE));
	rgbColorRGBA = *pColor;
	return SUCCEEDED(hr);
}

///////////////////////////////////////////////////////////////////////////////
// Function name	: XFile::GetFaceArray
// Description	    : Gets an array of faces
// Return type		: bool 
// Argument         : LPCSTR szName
// Argument         : meshface *&pFaces
//
bool XFile::GetFaceArray(LPCSTR szName, meshface *&pFaces)
{
	DWORD	dwSize;
	HRESULT	hr = m_pCurrentObject->GetData(szName, &dwSize,
										reinterpret_cast<void**>(&pFaces));
	PEASSERT(SUCCEEDED(hr) && 0 == dwSize % sizeof(meshface));
	return SUCCEEDED(hr);
}

///////////////////////////////////////////////////////////////////////////////
// Function name	: XFile::Get2DTextureCoordArray
// Description	    : Gets an array of 2D texture coords
// Return type		: bool 
// Argument         : LPCSTR szName
// Argument         : coords2d *&pCoords
//
bool XFile::Get2DTextureCoordArray(LPCSTR szName, coords2d *&pCoords)
{
	DWORD	dwSize;
	HRESULT	hr = m_pCurrentObject->GetData(szName, &dwSize,
										reinterpret_cast<void**>(&pCoords));
	PEASSERT(SUCCEEDED(hr) && 0 == dwSize % sizeof(coords2d));
	return SUCCEEDED(hr);
}

///////////////////////////////////////////////////////////////////////////////
// Function name	: XFile::GetChildDataObject
// Description	    : Gets the next child data object
// Return type		: bool 
// Argument         : IDirectXFileDataPtr &pDataObject
// Argument         : const IDirectXFileDataPtr & pFileObject
//
bool XFile::GetChildDataObject(IDirectXFileDataPtr &pDataObject,
							   const IDirectXFileDataPtr & pFileObject)
{
	HRESULT	hr;
	IDirectXFileObjectPtr	pChildObject;

	if(SUCCEEDED(hr = pFileObject->GetNextObject(&pChildObject.Interface())))
	{
		if(FAILED(hr = pChildObject.QueryInterface(pDataObject)))
		{
			IDirectXFileDataReferencePtr	pReference;
			if(SUCCEEDED(hr = pChildObject.QueryInterface(pReference)))
			{
				D3DEX(pReference->Resolve(&pDataObject.Interface()));
				pReference = NULL;
			}
			else
				PEASSERT(false);
		}
	}
	return SUCCEEDED(hr);
}
